// Shader.cpp: Implementierung der Klasse Shader
//
//////////////////////////////////////////////////////////////////////


#include <iostream>
#include <fstream>
//#include<gl/gl.h>
//#include<gl/glu.h>
#include <gl/glut.h>
#include <gl/glext.h>
#include "Shader.h"

using namespace std;

char* vs_name = "GLSlangBasicVS.glsl";
char* ps_name = "GLSlangBasicPS.glsl";

bool loaded;
/* glsl objects */
GLhandleARB _glVertexShader;
GLhandleARB _glPixelShader;

// glsl functions
PFNGLCREATEPROGRAMOBJECTARBPROC glCreateProgramObjectARB = NULL;
PFNGLCREATESHADEROBJECTARBPROC glCreateShaderObjectARB = NULL;
PFNGLLINKPROGRAMARBPROC glLinkProgramARB = NULL;
PFNGLCOMPILESHADERARBPROC glCompileShaderARB = NULL;
PFNGLGETINFOLOGARBPROC glGetInfoLogARB = NULL;
PFNGLDELETEOBJECTARBPROC glDeleteObjectARB = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC glUseProgramObjectARB = NULL;
PFNGLSHADERSOURCEARBPROC glShaderSourceARB = NULL;
PFNGLATTACHOBJECTARBPROC glAttachObjectARB = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC glGetObjectParameterivARB = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC glGetUniformLocationARB = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC glGetAttribLocationARB = NULL;
PFNGLUNIFORMMATRIX4FVARBPROC glUniformMatrix4fvARB;
PFNGLUNIFORM4FARBPROC glUniform4fARB = NULL;
PFNGLUNIFORM2FARBPROC glUniform2fARB = NULL;
PFNGLUNIFORM1FARBPROC glUniform1fARB = NULL;
PFNGLUNIFORM1IARBPROC glUniform1iARB = NULL;


bool InitializeGLSL()
{
	std::cout << "Shader: checking extensions...\n";
	
	char* extensions = (char*)glGetString(GL_EXTENSIONS);
	if(strstr(extensions, "GL_ARB_shading_language_100")==NULL) return false;
	if(strstr(extensions, "GL_ARB_shader_objects")==NULL) return false;
	
	std::cout << "Shader: ...extensions checked.\n";

	glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)wglGetProcAddress("glCreateProgramObjectARB");
	glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)wglGetProcAddress("glCreateShaderObjectARB");
	glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)wglGetProcAddress("glCompileShaderARB");
	glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)wglGetProcAddress("glLinkProgramARB");
	glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)wglGetProcAddress("glGetInfoLogARB");
	glDeleteObjectARB  = (PFNGLDELETEOBJECTARBPROC)wglGetProcAddress("glDeleteObjectARB");
	glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)wglGetProcAddress("glUseProgramObjectARB");
	glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)wglGetProcAddress("glShaderSourceARB");
	glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)wglGetProcAddress("glAttachObjectARB");
	glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)wglGetProcAddress("glGetObjectParameterivARB");
	glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)wglGetProcAddress("glGetUniformLocationARB");
	glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC)wglGetProcAddress("glGetAttribLocationARB");
	glUniformMatrix4fvARB = (PFNGLUNIFORMMATRIX4FVARBPROC)wglGetProcAddress("glUniformMatrix4fvARB");
	glUniform4fARB = (PFNGLUNIFORM4FARBPROC)wglGetProcAddress("glUniform4fARB");
	glUniform2fARB = (PFNGLUNIFORM2FARBPROC)wglGetProcAddress("glUniform2fARB");
	glUniform1fARB = (PFNGLUNIFORM1FARBPROC)wglGetProcAddress("glUniform1fARB");
	glUniform1iARB = (PFNGLUNIFORM1IARBPROC)wglGetProcAddress("glUniform1iARB");

   if(!glCreateProgramObjectARB || !glCreateShaderObjectARB ||  !glCompileShaderARB ||
      !glLinkProgramARB || !glGetInfoLogARB || !glDeleteObjectARB || !glUseProgramObjectARB ||
      !glShaderSourceARB || !glAttachObjectARB || !glGetObjectParameterivARB ||
      !glGetUniformLocationARB || !glUniform4fARB || !glUniform2fARB ||
      !glUniform1iARB || !glUniform1fARB)
      {
         return false;
      }

   return true;
}

/* load desired shaders using GLSlang*/
bool LoadShaders(char* vertexShader, char* pixelShader, GLhandleARB* glContext)
{ 
	// glCreateProgramObjectARB() will create a program object.  A program object is
    // a container that can hold multiple shaders in it.
    *glContext = glCreateProgramObjectARB();

    char error[4096];
    int result;

    // We use our loadShader to read in the entire file into ShaderCode.
    char *ShaderCode = loadShader(vs_name);

    // glCreateShaderObjectARB() will create a vertex or pixel shader.
    // Here we are sending it GL_VERTEX_SHADER_ARB to tell it we want a vertex shader.
    _glVertexShader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
   
    // glShaderSourceARB() will take our shader code and load it into our shader
    // object.  Here we send the vertex shader, number of strings, array of strings,
    // and an array of string lenghts.
    glShaderSourceARB(_glVertexShader, 1, (const char**)&ShaderCode, NULL);
   
    // glCompileShaderARB() will take our code and compile it down for GPU use.
    glCompileShaderARB(_glVertexShader);
   
    // glGetObjectParameterivARB() will get the results of anything we ask for
    // here we want the results of the compile status done on the vertex shader
    // this result is stored in result
    glGetObjectParameterivARB(_glVertexShader, GL_OBJECT_COMPILE_STATUS_ARB, &result);

    // since loadShader allocated memory for the vertex code, here we will delete it
    delete[] ShaderCode;
    ShaderCode = NULL;

	// if the result was not successful then we can print out an error
    if(!result)
    {
		// get the error message
        glGetInfoLogARB(_glVertexShader, sizeof(error), NULL, error); 
        // display the error to the user
        printf("Compile Error in Vertex Shader, error description:\n");
		printf(error);
		return false;
	}
   


    // if we made it here then the vertex shader compiled successfully
    // we can attach it to our program object using glAttachObjectARB()
    glAttachObjectARB(*glContext, _glVertexShader);

    // next we load the fragement (pixel) shader
    ShaderCode = loadShader(ps_name);

    _glPixelShader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
    glShaderSourceARB(_glPixelShader, 1, (const char**)&ShaderCode, NULL);
    glCompileShaderARB(_glPixelShader);
    glGetObjectParameterivARB(_glPixelShader, GL_OBJECT_COMPILE_STATUS_ARB, &result);

    // since loadShader allocated memory for the vertex code, here we will delete it
    delete[] ShaderCode;
    ShaderCode = NULL;

    // display any errors
    if(!result)
	{
		glGetInfoLogARB(_glPixelShader, sizeof(error), NULL, error);
        printf("Compile Error in Pixel Shader, error description:\n");
		printf(error);
		return false;
	}

    // attach this to the program object along with the vertex shader
    glAttachObjectARB(*glContext, _glPixelShader);


	// glLinkProgramARB() will link all the shaders in the program object
	// If there was a problem with any of them, we can display the error
	glLinkProgramARB(*glContext);
	glGetObjectParameterivARB(*glContext, GL_OBJECT_LINK_STATUS_ARB, &result);
	if(!result)
	{
		glGetInfoLogARB(*glContext, sizeof(error), NULL, error);
        printf("Error linking shaders, error description:\n");
		printf(error);
		return false;
	}

	return true;
}


char* loadShader(char *shaderName)
{
   char tempLine[256] = {0};
   int count = 0;

   // input object.
   ifstream tempInput;

   // Open the shader file.
   tempInput.open(shaderName);

   // If there is any errors then return false.
   if(!tempInput.is_open())
      {
         return NULL;
      }


   // Loop through each line of the file and get the total size.
   while(!tempInput.eof())
      {
         tempInput.getline(tempLine, 256, '\n');
         count++;
      }

   // Close the shader.
   tempInput.close();

   if(count == 0) return NULL;

   // Read in the data for use this time.
   ifstream input;

   // Create array to hold shader code.
   char *ShaderSource = new char[256 * count];

   // Re-open the shader and read in the whole thing into the array.
   input.open(shaderName);
   input.getline(ShaderSource , 256 * count, '\0');

   input.close(); // Close the shader.

   return ShaderSource;
}


void printInfoLog(GLhandleARB glContext)
{
    int maxLength = 0;
    glGetObjectParameterivARB(glContext, GL_OBJECT_INFO_LOG_LENGTH_ARB, &maxLength);

    char *infoLog = new char[maxLength];
    glGetInfoLogARB(glContext, maxLength, &maxLength, infoLog);

    cout << infoLog << endl;
}

bool isLoaded()
{
	return loaded;
}


void clearShaders()
{
	if(loaded)
	{
		glDeleteObjectARB(_glVertexShader);
		glDeleteObjectARB(_glPixelShader);
	}
}
