// Wall.cpp: Implementierung der Klasse Wall
//
///////////////////////////////////////////////////////////////////////

#include <GL/glut.h>
#include <cstdlib>
#include <iostream>

#include "Cg_Shader.h"
#include "Wall.h"


Wall::Wall(int _width, int _height, float _spacing)
{

	width = _width;
	height = _height;
	spacing = _spacing;
	// the +1 is a hack
	numVerticesX = (int)(ceil)(width/spacing)+1;
	//std::cout << numVerticesX << '\n';
	// the +2 is a hack
	numVerticesY = (int)(ceil)(height/spacing)+2;
	//std::cout << numVerticesY << '\n';

	// allocate memory
	vertices = (Vector4*) malloc(numVerticesX*numVerticesY*sizeof(Vector4));
	objectToTangentSpaceMatrices = (Matrix4x4*) malloc(numVerticesX*numVerticesY*sizeof(Matrix4x4));
	//initialize grid
	initGrid();

}

void Wall::LoadDetail(char* textureName)
{
	texture.LoadDetail(textureName);
}

void Wall::LoadBumpmap(char* bumpmapName)
{
	texture.LoadBumpmap(bumpmapName);
}

void Wall::initGrid()
{
	
	for (int x = 0; x < numVerticesX; x++) 
	{	
		
		for (int y = 0; y < numVerticesY; y++) 
		{

			vertices[x*numVerticesY + y].x = x*spacing;
			vertices[x*numVerticesY + y].y = y*spacing;
			vertices[x*numVerticesY + y].z = 0.0f;
		
			// there are four cases to be distinguished
			if(y>0&&x>0) // general case
			{
				Vector4 v0 = vertices[x*numVerticesY + y];
				Vector4 v1 = vertices[(x-1)*numVerticesY + y];
				Vector4 v2 = vertices[x*numVerticesY + y - 1];
				float s0 = vertices[x*numVerticesY + y].x/width;
				float t0 = vertices[x*numVerticesY + y].y/height;
				float s1 = vertices[(x-1)*numVerticesY + y].x/width;
				float t1 = vertices[(x-1)*numVerticesY + y].y/height;
				float s2 = vertices[x*numVerticesY + y - 1].x/width;
				float t2 = vertices[x*numVerticesY + y - 1].y/height;
			
				// for each vertex...
				Vector4 dv1 = v1-v0;
				Vector4 dv2 = v2-v0;
				float ds1 = s1-s0;
				float dt1 = t1-t0;
				float ds2 = s2-s0;
				float dt2 = t2-t0;
				float denominator = ds1*dt2 - ds2*dt1;
				float scale = 1.0f/denominator;
				Vector4 T = Vector4((dt2*dv1.x - dt1*dv2.x)/scale,
									(dt2*dv1.y - dt1*dv2.y)/scale,
									(dt2*dv1.z - dt1*dv2.z)/scale);
				T.Normalize();
				Vector4 B = Vector4((-ds2*dv1.x - ds1*dv2.x)/scale,
									(-ds2*dv1.y - ds1*dv2.y)/scale,
									(-ds2*dv1.z - ds1*dv2.z)/scale);
				B.Normalize();
				Vector4 N;
				N.CrossProduct(T,B);
				Matrix4x4 BumpMatrix = Matrix4x4(T,B,N);
				BumpMatrix.inverseMatrix(BumpMatrix);
				objectToTangentSpaceMatrices[x*numVerticesY + y] = BumpMatrix;

			}
			else if(y==0&&x>0) // first row
			{

				Vector4 v0 = vertices[x*numVerticesY];
				Vector4 v1 = vertices[(x-1)*numVerticesY];
				Vector4 v2 = vertices[(x-1)*numVerticesY+1];
				float s0 = vertices[x*numVerticesY].x/width;
				float t0 = vertices[x*numVerticesY].y/height;
				float s1 = vertices[(x-1)*numVerticesY].x/width;
				float t1 = vertices[(x-1)*numVerticesY].y/height;
				float s2 = vertices[(x-1)*numVerticesY+1].x/width;
				float t2 = vertices[(x-1)*numVerticesY+1].y/height;
							
				Vector4 dv1 = v1-v0;
				Vector4 dv2 = v2-v0;
				float ds1 = s1-s0;
				float dt1 = t1-t0;
				float ds2 = s2-s0;
				float dt2 = t2-t0;
				float denominator = ds1*dt2 - ds2*dt1;
				float scale = 1.0f/denominator;
				Vector4 T = Vector4((dt2*dv1.x - dt1*dv2.x)/scale,
									(dt2*dv1.y - dt1*dv2.y)/scale,
									(dt2*dv1.z - dt1*dv2.z)/scale);
				T.Normalize();
				Vector4 B = Vector4((-ds2*dv1.x - ds1*dv2.x)/scale,
									(-ds2*dv1.y - ds1*dv2.y)/scale,
									(-ds2*dv1.z - ds1*dv2.z)/scale);
				B.Normalize();
				Vector4 N;
				N.CrossProduct(T,B);
				Matrix4x4 BumpMatrix = Matrix4x4(T,B,N);
				BumpMatrix.inverseMatrix(BumpMatrix);
				objectToTangentSpaceMatrices[x*numVerticesY + y] = BumpMatrix;

			}else if(y>0&&x==0)	// first column
			{

				Vector4 v0 = vertices[y];
				Vector4 v1 = Vector4(spacing, (y-1)*spacing, 0.0f);
				Vector4 v2 = vertices[y-1];
				float s0 = vertices[y].x/width;
				float t0 = vertices[y].y/height;
				float s1 = spacing;
				float t1 = (y-1)*spacing;
				float s2 = vertices[y-1].x/width;
				float t2 = vertices[y-1].y/height;
			
				Vector4 dv1 = v1-v0;
				Vector4 dv2 = v2-v0;
				float ds1 = s1-s0;
				float dt1 = t1-t0;
				float ds2 = s2-s0;
				float dt2 = t2-t0;
				float denominator = ds1*dt2 - ds2*dt1;
				float scale = 1.0f/denominator;
				Vector4 T = Vector4((dt2*dv1.x - dt1*dv2.x)/scale,
									(dt2*dv1.y - dt1*dv2.y)/scale,
									(dt2*dv1.z - dt1*dv2.z)/scale);
				T.Normalize();
				Vector4 B = Vector4((-ds2*dv1.x - ds1*dv2.x)/scale,
									(-ds2*dv1.y - ds1*dv2.y)/scale,
									(-ds2*dv1.z - ds1*dv2.z)/scale);
				B.Normalize();
				Vector4 N;
				N.CrossProduct(T,B);
				Matrix4x4 BumpMatrix = Matrix4x4(T,B,N);
				BumpMatrix.inverseMatrix(BumpMatrix);
				objectToTangentSpaceMatrices[x*numVerticesY + y] = BumpMatrix;

			}
			else if(y==0&&x==0)	// first element
			{

				Vector4 v0 = vertices[0];
				Vector4 v1 = Vector4(spacing, 0.0f, 0.0f);
				Vector4 v2 = Vector4(0.0f, spacing, 0.0f);
				float s0 = vertices[0].x/width;
				float t0 = vertices[0].y/height;
				float s1 = spacing;
				float t1 = 0.0f;
				float s2 = 0.0f;
				float t2 = spacing;
			
				Vector4 dv1 = v1-v0;
				Vector4 dv2 = v2-v0;
				float ds1 = s1-s0;
				float dt1 = t1-t0;
				float ds2 = s2-s0;
				float dt2 = t2-t0;
				float denominator = ds1*dt2 - ds2*dt1;
				float scale = 1.0f/denominator;
				Vector4 T = Vector4((dt2*dv1.x - dt1*dv2.x)/scale,
									(dt2*dv1.y - dt1*dv2.y)/scale,
									(dt2*dv1.z - dt1*dv2.z)/scale);
				T.Normalize();
				Vector4 B = Vector4((-ds2*dv1.x - ds1*dv2.x)/scale,
									(-ds2*dv1.y - ds1*dv2.y)/scale,
									(-ds2*dv1.z - ds1*dv2.z)/scale);
				B.Normalize();
				Vector4 N;
				N.CrossProduct(T,B);
				Matrix4x4 BumpMatrix = Matrix4x4(T,B,N);
				BumpMatrix.inverseMatrix(BumpMatrix);
				objectToTangentSpaceMatrices[x*numVerticesY + y] = BumpMatrix;

			}

		}
	}		

}

void Wall::draw() 
{

	float col[] = {0.3f, 0.2f, 0.8f};
	glColor3fv(col);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_DST_COLOR);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE,col);
	glBindTexture(GL_TEXTURE_2D, texture.detailHandle);
	glNormal3f(0.0f, 0.0f, 1.0f);
	
	for(int x=0; x < numVerticesX-1; x++)
	{
		glBegin(GL_TRIANGLE_STRIP);
		for(int y=0; y < numVerticesY-1; y++)
		{	
			glTexCoord2f(vertices[x*numVerticesY+y].x/width,vertices[x*numVerticesY+y].y/height);
			glVertex3f(vertices[x*numVerticesY+y].x,vertices[x*numVerticesY+y].y,vertices[x*numVerticesY+y].z);
			glTexCoord2f(vertices[(x+1)*numVerticesY+y].x/width,vertices[(x+1)*numVerticesY+y].y/height);
			glVertex3f(vertices[(x+1)*numVerticesY+y].x,vertices[(x+1)*numVerticesY+y].y,vertices[(x+1)*numVerticesY+y].z);
		}
		glEnd();
	}
	glDisable(GL_BLEND);

}

/*void Wall::drawBump(Vector4 light, Camera camera)
{

	for(int x=0; x < numVerticesX-1; x++)
	{
		glBegin(GL_TRIANGLE_STRIP);
		for(int y=0; y < numVerticesY-1; y++)
		{	

			Matrix4x4 currentObjectToTangentSpaceMatrix = objectToTangentSpaceMatrices[x*numVerticesY + y];
			Vector4 currentVertex = vertices[x*numVerticesY + y];
			Vector4 vertexToLight = light-currentVertex;
			vertexToLight = currentObjectToTangentSpaceMatrix.VectorMatrixMultiply(vertexToLight);
			vertexToLight.Normalize();
			vertexToLight.RangeCompress(1.0f);
			// set base color
			glColor3f(vertexToLight.x, vertexToLight.y, vertexToLight.z);
			Vector4 view = Vector4(camera.xPos-currentVertex.x, camera.yPos-currentVertex.y, camera.zPos-currentVertex.z);
			Vector4 halfVector = light + view;
			halfVector = currentObjectToTangentSpaceMatrix.VectorMatrixMultiply(halfVector);
			halfVector.Normalize();
			halfVector.RangeCompress(1.0f);
			// set secondary color
			glSecondaryColor3fEXT(halfVector.x, halfVector.y, halfVector.z);
			// set texture and vertex attributes
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, currentVertex.x/width, currentVertex.y/height);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB, currentVertex.x/width, currentVertex.y/height);
			glVertex3f(currentVertex.x,currentVertex.y,currentVertex.z);

			currentObjectToTangentSpaceMatrix = objectToTangentSpaceMatrices[(x+1)*numVerticesY + y];
			currentVertex = vertices[(x+1)*numVerticesY + y];
			vertexToLight = light-currentVertex;
			vertexToLight = currentObjectToTangentSpaceMatrix.VectorMatrixMultiply(vertexToLight);
			vertexToLight.Normalize();
			vertexToLight.RangeCompress(1.0f);
			// set base color
			glColor3f(vertexToLight.x, vertexToLight.y, vertexToLight.z);
			view = Vector4(camera.xPos-currentVertex.x, camera.yPos-currentVertex.y, camera.zPos-currentVertex.z);
			halfVector = light + view;
			halfVector = currentObjectToTangentSpaceMatrix.VectorMatrixMultiply(halfVector);
			halfVector.Normalize();
			halfVector.RangeCompress(1.0f);
			// set secondary color
			glSecondaryColor3fEXT(halfVector.x, halfVector.y, halfVector.z);
			// set texture and vertex attributes
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, currentVertex.x/width, currentVertex.y/height);
			glMultiTexCoord2fARB(GL_TEXTURE1_ARB, currentVertex.x/width, currentVertex.y/height);
			glVertex3f(currentVertex.x,currentVertex.y,currentVertex.z);

		}
		glEnd();
	}

}*/

void Wall::drawBump()
{
	// simplified version to speed up application
	glBegin(GL_QUADS);

		// lower left corner
		// set texture and vertex attributes for both the detail and the normal map
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 0.0f);
		// specify the tangent matrix vectors one by one and send them to the vertex shader
		glMultiTexCoord3fARB(GL_TEXTURE1_ARB, 1.0f, 0.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE2_ARB, 0.0f, 1.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE3_ARB, 0.0f, 0.0f, 1.0f);
		// finally set the vertex coordinates
		glVertex3f(0.0f, 0.0f, 0.0f);

		// upper left corner
		// set texture and vertex attributes for both the detail and the normal map
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0.0f, 1.0f);
		// specify the tangent matrix vectors one by one and send them to the vertex shader
		glMultiTexCoord3fARB(GL_TEXTURE1_ARB, 1.0f, 0.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE2_ARB, 0.0f, 1.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE3_ARB, 0.0f, 0.0f, 1.0f);
		// finally set the vertex coordinates
		glVertex3f(0.0f, height, 0.0f);
	
		// upper right corner
		// set texture and vertex attributes for both the detail and the normal map
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 1.0f);
		// specify the tangent matrix vectors one by one and send them to the vertex shader
		glMultiTexCoord3fARB(GL_TEXTURE1_ARB, 1.0f, 0.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE2_ARB, 0.0f, 1.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE3_ARB, 0.0f, 0.0f, 1.0f);
		// finally set the vertex coordinates
		glVertex3f(width, height, 0.0f);

		// lower right corner
		// set texture and vertex attributes for both the detail and the normal map
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1.0f, 0.0f);
		// specify the tangent matrix vectors one by one and send them to the vertex shader
		glMultiTexCoord3fARB(GL_TEXTURE1_ARB, 1.0f, 0.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE2_ARB, 0.0f, 1.0f, 0.0f);
		glMultiTexCoord3fARB(GL_TEXTURE3_ARB, 0.0f, 0.0f, 1.0f);
		// finally set the vertex coordinates
		glVertex3f(width, 0.0f, 0.0f);

	glEnd();

	/*
	// computationally more complicated version
	for(int x=0; x < numVerticesX-1; x++)
	{
		glBegin(GL_TRIANGLE_STRIP);
		for(int y=0; y < numVerticesY-1; y++)
		{	

			// get the current object to tangent space matrix
			Matrix4x4 currentMatrix = objectToTangentSpaceMatrices[x*numVerticesY + y];
			// get the current vertex
			Vector4 currentVertex = vertices[x*numVerticesY + y];
			// set texture and vertex attributes for both the detail and the normal map
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, currentVertex.x/width, currentVertex.y/height);
			// specify the tangent matrix vectors one by one and send them to the vertex shader
			glMultiTexCoord3fARB(GL_TEXTURE1_ARB, currentMatrix.matrix[0], currentMatrix.matrix[4], currentMatrix.matrix[8]);
			glMultiTexCoord3fARB(GL_TEXTURE2_ARB, currentMatrix.matrix[1], currentMatrix.matrix[5], currentMatrix.matrix[9]);
			glMultiTexCoord3fARB(GL_TEXTURE3_ARB, currentMatrix.matrix[2], currentMatrix.matrix[6], currentMatrix.matrix[10]);
			// finally set the vertex coordinates
			glVertex3f(currentVertex.x,currentVertex.y,currentVertex.z);


			// get the current object to tangent space matrix
			currentMatrix = objectToTangentSpaceMatrices[(x+1)*numVerticesY + y];
			// get the current vertex
			currentVertex = vertices[(x+1)*numVerticesY + y];
			// set texture and vertex attributes for both the detail and the normal map
			glMultiTexCoord2fARB(GL_TEXTURE0_ARB, currentVertex.x/width, currentVertex.y/height);
			// specify the tangent matrix vectors one by one and send them to the vertex shader
			glMultiTexCoord3fARB(GL_TEXTURE1_ARB, currentMatrix.matrix[0], currentMatrix.matrix[4], currentMatrix.matrix[8]);
			glMultiTexCoord3fARB(GL_TEXTURE2_ARB, currentMatrix.matrix[1], currentMatrix.matrix[5], currentMatrix.matrix[9]);
			glMultiTexCoord3fARB(GL_TEXTURE3_ARB, currentMatrix.matrix[2], currentMatrix.matrix[6], currentMatrix.matrix[10]);
			// finally set the vertex coordinates
			glVertex3f(currentVertex.x,currentVertex.y,currentVertex.z);

		}
		glEnd();
	}
	*/
}

float Wall::getWidth()
{
	return width;
}

float Wall::getHeight()
{
	return height;
}

float Wall::getNumVerticesX() 
{ 
	return numVerticesX;
}

float Wall::getNumVerticesY() 
{ 
	return numVerticesY;
}

float Wall::getSpacing() 
{ 
	return spacing;
}

unsigned int Wall::getDetailHandle()
{
	return texture.detailHandle;
}

unsigned int Wall::getBumpHandle()
{
	return texture.bumpmapHandle;
}

Vector4* Wall::getVertices() 
{ 
	return vertices;
}